/************************************************************************/
/*   This file is a part of the mkprom3 boot-prom utility               */
/*   Copyright (C) 2004 Gaisler Research                                */
/*                                                                      */
/*   This library is free software; you can redistribute it and/or      */
/*   modify it under the terms of the GNU General Public                */
/*   License as published by the Free Software Foundation; either       */
/*   version 2 of the License, or (at your option) any later version.   */
/*                                                                      */
/*   See the file COPYING.GPL for the full details of the license.      */
/************************************************************************/

	
/*
 * locore.s for LEON3
 *
 */
 
/* The traptable has to be the first code in a boot PROM. */

/* Entry for traps which jump to a programmer-specified trap handler.  */
#define TRAPR(H)  wr %g0, 0x0e0, %psr; mov %g0, %tbr; ba (H); mov %g0, %wim;
#define TRAP(H)  mov %psr, %l0; sethi %hi(H), %l4; jmp %l4+%lo(H); nop;

/* Unexcpected trap will halt the processor by forcing it to error state */
#define BAD_TRAP ta 0; nop; nop; nop;

/* Number of register windows */
#define NWINDOWS 	8

#define WIM_INIT   2

  .seg    "text"
  .global _trap_table, start, _memcfg1, _memcfg2, _stack, _ftreg 
  .global bdinit1, bdinit2, main, getpsr, crt2

start:
_trap_table:
  TRAPR(_hardreset);		! 00 reset trap 
  BAD_TRAP;      		! 01 instruction_access_exception
  BAD_TRAP;      		! 02 illegal_instruction
  BAD_TRAP;      		! 03 priveleged_instruction
  TRAP(_fpu_dis);		! 04 fp_disabled
  TRAP(_window_overflow);	! 05 window_overflow
  TRAP(_window_underflow);	! 06 window_underflow
  BAD_TRAP;      		! 07 memory_address_not_aligned
  BAD_TRAP;      		! 08 fp_exception
  BAD_TRAP;      		! 09 data_access_exception
  BAD_TRAP;      		! 0A tag_overflow
  BAD_TRAP;			! 0B undefined
  BAD_TRAP;			! 0C undefined
  BAD_TRAP;			! 0D undefined
  BAD_TRAP;			! 0E undefined
  BAD_TRAP;			! 0F undefined
  BAD_TRAP;			! 10 undefined
  TRAP(_irqhdl);      		! 11 interrupt level 1
  TRAP(_irqhdl);      		! 12 interrupt level 2
  TRAP(_irqhdl);      		! 13 interrupt level 3
  TRAP(_irqhdl);      		! 14 interrupt level 4
  TRAP(_irqhdl);      		! 15 interrupt level 5
  TRAP(_irqhdl);      		! 16 interrupt level 6
  TRAP(_irqhdl);      		! 17 interrupt level 7
  TRAP(_irqhdl);      		! 18 interrupt level 8
  TRAP(_irqhdl);      		! 19 interrupt level 9
  TRAP(_irqhdl);      		! 1A interrupt level 1
  TRAP(_irqhdl);      		! 1B interrupt level 11
  TRAP(_irqhdl);      		! 1C interrupt level 12
  TRAP(_irqhdl);      		! 1D interrupt level 13
  TRAP(_irqhdl);      		! 1E interrupt level 14
  TRAP(_irqhdl);      		! 1F interrupt level 15

_hardreset:

	flush
	clr	%g1
	clr	%g2
	clr	%g3
	clr	%g4
	clr	%g5
	clr	%g6
	clr	%g7

	set	noinit, %g2		! Load memory config register 1
	ld	[%g2], %g2
	subcc   %g2, %g0, %g0
	bne	sysini0

	set	0x81000f, %g2
	mov	%psr, %g1
	srl	%g1, 24, %g1
	and  	%g1, 0x0f, %g1
	subcc	%g1, 3, %g1
	be	1f
	set     0x80000000, %g1
  	st	%g2, [%g1+0x14]
	set	0xAA00, %g3
	ba 2f
  	st	%g3, [%g1+0xA4]
1:
	set	_memcaddr, %g1
	ld	[%g1], %g1		
  	sta	%g2, [%g0] 2	        ! enable LEON3 cache
        st      %g0, [%g1 + 0x08]     
2:
	set	_memcfg1, %g2		! Load memory config register 1
	ld	[%g2], %g2
  	ld	[%g1], %g3		! 
	and	%g3, 0x300, %g3
	or	%g2, %g3, %g2
  	st	%g2, [%g1 + 0x00]
	set	_memcfg2, %g2		! Load memory config register 2
	ld	[%g2], %g2
  	st	%g2, [%g1 + 0x04]
	set	_memcfg3, %g2		! Load memory config register 3
	ld	[%g2], %g2

	set	freq, %g2		! Load timer scaler register
	ld	[%g2], %g2
	be	1f
	sub	%g2, 1, %g2

	st	%g2, [%g1 + 0x64]
	st	%g2, [%g1 + 0x60]
	set	-1, %g3
	st	%g3, [%g1 + 0x40]
	st	%g3, [%g1 + 0x44]
	set	7, %g3
	st	%g3, [%g1 + 0x48]
	set	0x80000070, %g1
	ba	2f
	nop
1:
	set	_gptaddr, %g1
	ld	[%g1], %g1	
  	st	%g2, [%g1 + 0x4]
  	st	%g2, [%g1 + 0x0]
	set	-1, %g3
	st	%g3, [%g1 + 0x10]
	st	%g3, [%g1 + 0x14]
	set	7, %g3
	st	%g3, [%g1 + 0x18]

	set	_uaddr, %g1
	ld	[%g1], %g1	
2:
	set	_uart, %g2		! Load UART scaler register
	ld	[%g2], %g2
  	st	%g2, [%g1 + 0xc]
	set	3, %g2			! Enable UART
  	st	%g2, [%g1 + 0x8]
	st	%g0, [%g1 + 0x4]

	mov	%g0, %y 		! Init Y-register

	set	NWINDOWS, %g2
wl0:
	clr	%i0
	clr	%i1
	clr	%i2
	clr	%i3
	clr	%i4
	clr	%i5
	clr	%i6
	clr	%i7
	clr	%l0
	clr	%l1
	clr	%l2
	clr	%l3
	clr	%l4
	clr	%l5
	clr	%l6
	clr	%l7
	save
	subcc	%g2, 1, %g2
	bne	wl0
	nop

sysini0:
	set	WIM_INIT, %g3
	mov	%g3, %wim
	set	0x1000, %g3
	mov	%psr, %g2
	wr	%g2, %g3, %psr
	nop; nop; nop
	set	ifsr, %g1
	ld	[%g1], %fsr		! Check if FPU is present

	ba	fixfpu
	nop

! FPU disabled trap address

_fpu_dis:
	mov	%g2, %psr
	nop; nop; nop
	ba    sysini
	nop
	

! Wash FPU register file

fixfpu:

	ldd	[%g0], %f0
	ldd	[%g0], %f2
	ldd	[%g0], %f4
	ldd	[%g0], %f6
	ldd	[%g0], %f8
	ldd	[%g0], %f10
	ldd	[%g0], %f12
	ldd	[%g0], %f14
	ldd	[%g0], %f16
	ldd	[%g0], %f18
	ldd	[%g0], %f20
	ldd	[%g0], %f22
	ldd	[%g0], %f24
	ldd	[%g0], %f26
	ldd	[%g0], %f28
	ldd	[%g0], %f30

! System init

sysini:

	call	bdinit1
	nop

	set	_stack, %g2
	ld	[%g2], %fp
	andn	%fp, 0x0f, %fp
	sub	%fp, 64, %sp

        sub     %sp, 0x40, %sp  	! room for main to save args
	call	crt2
        nop

	mov	%psr, %g2
	wr 	%g2, 0x20, %psr
	nop; nop; nop
	ta 	0			! Halt if _main would return ...
	nop



!Window overflow trap handler.
  	.global _window_overflow

_window_overflow:

  	mov  %wim, %l3       		! Calculate next WIM
  	mov  %g1, %l7           
  	srl  %l3, 1, %g1        
  	sll  %l3, NWINDOWS-1 , %l4
  	or   %l4, %g1, %g1

  	save                              ! Get into window to be saved.
  	mov  %g1, %wim
  	nop; nop; nop
        st   %l0, [%sp + 0];
        st   %l1, [%sp + 4];
        st   %l2, [%sp + 8];
        st   %l3, [%sp + 12];
        st   %l4, [%sp + 16]; 
        st   %l5, [%sp + 20];
        st   %l6, [%sp + 24]; 
        st   %l7, [%sp + 28]; 
        st   %i0, [%sp + 32]; 
        st   %i1, [%sp + 36]; 
        st   %i2, [%sp + 40]; 
        st   %i3, [%sp + 44]; 
        st   %i4, [%sp + 48]; 
        st   %i5, [%sp + 52]; 
        st   %i6, [%sp + 56]; 
        st   %i7, [%sp + 60];
  	restore				! Go back to trap window.
  	mov  %l7, %g1
  	jmp  %l1			! Re-execute save.
  	rett %l2

  /* Window underflow trap handler.  */

  	.global  _window_underflow

_window_underflow:

	mov  %wim, %l3			! Calculate next WIM
	sll  %l3, 1, %l4
  	srl  %l3, NWINDOWS-1, %l5
  	or   %l5, %l4, %l5
  	mov  %l5, %wim
  	nop; nop; nop
  	restore				! Two restores to get into the
  	restore				! window to restore
        ld   [%sp + 0], %l0; 		! Restore window from the stack
        ld   [%sp + 4], %l1; 
        ld   [%sp + 8], %l2; 
        ld   [%sp + 12], %l3; 
        ld   [%sp + 16], %l4; 
        ld   [%sp + 20], %l5; 
        ld   [%sp + 24], %l6; 
        ld   [%sp + 28], %l7; 
        ld   [%sp + 32], %i0; 
        ld   [%sp + 36], %i1; 
        ld   [%sp + 40], %i2; 
        ld   [%sp + 44], %i3; 
        ld   [%sp + 48], %i4; 
        ld   [%sp + 52], %i5; 
        ld   [%sp + 56], %i6; 
        ld   [%sp + 60], %i7;
  	save				! Get back to the trap window.
  	save
  	jmp  %l1			! Re-execute restore.
  	rett  %l2

	.global sections

_irqhdl:		

	mov	%l1, %l5
	mov	%l2, %l6
	restore				! enter previous frame (cannot trap)
	mov	%psr, %g5
	or	%g5, 0x0f00, %g5
        wr      %g5, %psr		! disable interrupts
        wr      %g5, 0x20, %psr		! enable traps, disable interrupts
	mov	%tbr, %g6
	nop; nop
	save	%sp, -64, %sp		! save to flush window
	mov	%psr, %g5
	
	srl	%g6, 2, %g6
	and	%g6, 0x3c, %l7
	srl	%l7, 2, %o0
	set	_irqtbl, %l4
	ld	[%l4 + %l7], %l7
	mov	%g1, %l1
	mov	%g2, %l2
	mov	%g3, %l3
	mov	%g4, %l4
	jmpl	%l7, %o7
	mov	%y,  %g7

	restore
	save
	mov	%l1, %g1
	mov	%l2, %g2
	mov	%l3, %g3
	mov	%l4, %g4
	mov	%g7, %y

        wr      %g5, 0x20, %psr		! disable traps
	andn	%g5, 0x0f00, %l4
        wr      %l4, 0x20, %psr		! disable traps, enable interrupts
	nop; nop; nop
  	jmp  %l5			! Jump to PC
  	rett  %l6

getpsr:
	retl
	mov %psr, %o0

ifsr:
	.word 0
